#include <math.h>
#include <float.h>
#include <stdbool.h>
#include <stdio.h>
#include "gaussp.h"
#include "lustre.h"



long double RF(long double x, long double y, long double z)
{
 long double AmF = (x + y + z)/3, lambda, xm = x, ym = y, zm = z;
 long double X, Y, Z, E2, E3;
 long double AmxF = AmF - x, AmyF = AmF - y, AmzF = AmF - z, Q = ABS(AmxF) > ABS(AmyF) ? AmxF : AmyF;
 long double powinvfour = 1;
 Q = ABS(AmzF) > ABS(Q) ? ABS(AmzF) : ABS(Q);
 Q *= 0.8326831776556043 / RESID;
 do{
        lambda = sqrtl(xm * ym) + sqrtl(ym * zm) + sqrtl(xm * zm);
        AmF = (AmF + lambda) / 4;
        xm = (xm + lambda) / 4;
        ym = (ym + lambda) / 4;
        zm = (zm + lambda) / 4;
        powinvfour /= 4;
        Q /= 4;
 }while(Q > AmF);
 X = AmxF * powinvfour / AmF;
 Y = AmyF * powinvfour / AmF;
 Z = AmzF * powinvfour / AmF;
 E2 = X * Y - Z * Z;
 E3 = X * Y * Z;
 return 1 / sqrtl(AmF) * (1 - E2 / 10 + E3 / 14 + E2 * E2 / 24 - 3 * E2 * E3 / 44);
}

long double RJ(long double x, long double y, long double z, long double p)
{
 long double X, Y, Z, P, E2, E3, E4, E5, t;
 long double AmJ = (x + y + z + 2 * p)/5, xm = x, ym = y, zm = z, pm = p;
 long double AmxJ = AmJ - x, AmyJ = AmJ - y, AmzJ = AmJ - z, AmpJ = AmJ - p, Q = ABS(AmxJ) > ABS(AmyJ) ? AmxJ : AmyJ;
 long double sqx, sqy, sqz, sqp;
 long double sqdelta0 = (p - x) * (p - y) * (p - z);
 long double lambda, sqdelta8 = sqrtl(ABS(sqdelta0));
 int sgn = SIGN(sqdelta0);
 long double dm, powinvfour = 1, powertwo = (sgn < 0 ? 3 : 6) / (sgn == 0 ? 1 : sqdelta8), sumJ = 0;
 Q = ABS(AmzJ) > ABS(Q) ? ABS(AmzJ) : ABS(Q);
 Q = ABS(AmpJ) > Q ? ABS(AmpJ) : Q;
 Q *= 0.8908987181403393 / RESID;
 do{
        sqx = sqrtl(xm);  sqy = sqrtl(ym);  sqz = sqrtl(zm); sqp = sqrtl(pm);
        lambda = sqx * sqy + sqy * sqz + sqx * sqz;
        dm = (sqp + sqx) * (sqp + sqy) * (sqp + sqz);
        switch(sgn)
        {
         case 1:
           sumJ += powertwo * atan2l(sqdelta8, dm);
           break;
         case 0:
           sumJ += 6 * powinvfour / dm;
           break;
         default:
           t = 2 * sqdelta8 / (dm - sqdelta8);
           sumJ += (logl(1 + t)) * powertwo;           
        }
        AmJ = (AmJ + lambda) / 4;
        xm = (xm + lambda) / 4;
        ym = (ym + lambda) / 4;
        zm = (zm + lambda) / 4;
        pm = (pm + lambda) / 4;
        powinvfour /= 4;
        sqdelta8 /= 8;
        powertwo *= 2;
 }while(Q * powinvfour > AmJ);
 X = AmxJ * powinvfour / AmJ;
 Y = AmyJ * powinvfour / AmJ;
 Z = AmzJ * powinvfour / AmJ;
 P = (- X - Y - Z) / 2;
 E2 = X * Y + X * Z + Y * Z - 3 * P * P;
 E3 = X * Y * Z + 2 * E2 * P + 4 * P * P * P;
 E4 = (2 * X * Y * Z + E2 * P + 3 * P * P * P) * P;
 E5 = X * Y * Z * P * P;
 return sumJ + powinvfour / sqrtl(AmJ * AmJ * AmJ) * (1 - E2 * 3 / 14 + E3 / 6 + E2 * E2 * 9 / 88 - 3 * E4 / 22 -  9 * E2 * E3 / 52 + 3 * E5 / 26);
}

long double RD(long double x, long double y, long double z)
{
 long double AmD = (x + y + 3 * z)/5, lambda, xm = x, ym = y, zm = z;
 long double X, Y, Z, E2, E3, E4, E5;
 long double AmxD = AmD - x, AmyD = AmD - y, AmzD = AmD - z, Q = ABS(AmxD) > ABS(AmyD) ? AmxD : AmyD;
 long double powinvfour = 1, sumD = 0;
 Q = ABS(AmzD) > ABS(Q) ? ABS(AmzD) : ABS(Q);
 Q *= 0.8326831776556043l / RESID;
 do{
        lambda = sqrtl(xm * ym) + sqrtl(ym * zm) + sqrtl(xm * zm);
        sumD += powinvfour / (sqrtl(zm) * (zm + lambda));
        AmD = (AmD + lambda) / 4;
        xm = (xm + lambda) / 4;
        ym = (ym + lambda) / 4;
        zm = (zm + lambda) / 4;
        powinvfour /= 4;
        Q /= 4;
 }while(Q > AmD);
 X = AmxD * powinvfour / AmD;
 Y = AmyD * powinvfour / AmD;
 Z = -(X + Y) / 3;
 E2 = X * Y - 6 * Z * Z;
 E3 = (3 * X * Y - 8 * Z * Z) * Z;
 E4 = (X * Y - Z * Z) * Z * Z * 3;
 E5 = X * Y * Z * Z * Z;
 return 3 * sumD + powinvfour / sqrtl(AmD * AmD * AmD) * (1 - 3 * E2 / 14 + E3 / 6 + E2 * E2 * 9 / 88 - 3 * E4 / 22 - 9 / 52 * E2 * E3 + 3 / 26 * E5);
}

extern inline long double ReArcCos(long double x)
{
 if(x >= 1) return 0;
 if(x <= -1) return PI;
 return acosl(x);
}

long double Phi(long double delta, long double r)
{
 long double x;
 if(delta == 0 || r == 0) return 0;
 x = (delta * delta + r * r - 1) / delta / r / 2;
 if(x >= 1) return 1;
 if(x <= -1) return -1;
 return x;
}

long double L0(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi1 = (delta * delta + 1 - r * r) / 2;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
 {
  psi1 = (-SIGN(psi1) + 1) * PI_2;
  psi2 = (-SIGN(psi2) + 1) * PI_2;
 }
 else
 {
  psi1 = ReArcCos(psi1 / delta);
  psi2 = ReArcCos(psi2 / delta);
 }
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return psi1 + psi2 * r * r - Q / 2;
}

long double D1L0(long double delta, long double r)
{
 if(r <= 0 || delta <= 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  return 0;
 else
  Q = sqrtl(Q);
 return -Q / delta;
}

long double D2L0(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
  psi2 = (-2 * SIGN(psi2) + 1) * PI_2;
 else
  psi2 = ReArcCos(psi2 / delta);
 return 2 * r * psi2;
}


long double L1(long double delta, long double r)
{
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double rmindeltaq = rmindelta * rmindelta, rdelta4 = 4 * delta * r, k, n, s, c, P, F, E, res = 0, sqq;
 if(delta == 0.5L && r == 0.5L)
     return 0.602753106752153308376436683315L;
 if (ABS(rmindelta) >= 1 || r <= 0 || delta < 0)
 {
  if(r > delta) res += 2.0943951023931954923L;
  if(r == delta) res += 1.0471975511965977462L;
  return res;
 }
 k = rdelta4 / (1 - rmindeltaq);
 s = (k < 1) ? 1 : 1 / sqrtl(k);
 c = (k < 1) ? 1 : 1 / k;
 if(r == delta)
  P = 0;
 else
  {
   n = -rdelta4 / rmindeltaq;
   P = s * RF(ABS(1 - c), ABS(1 - c * k), 1) + s * c * n * RJ(ABS(1 - c), ABS(1 - c * k), 1, 1 - c * n) / 3;
   P *= 2 * rplusdelta;
   P /= - 3 * rmindelta;
  }
 F = s * RF(ABS(1 - c), ABS(1 - c * k), 1);
 F *= (- 3 * rmindelta * rplusdelta + (1 - rmindeltaq) * (rplusdelta * rplusdelta - 1)) * 2;
 F /= 9;
 res = P - F;
 sqq = sqrtl(1 - rmindeltaq);
 res /= sqq;
 E = s * RF(ABS(1 - c), ABS(1 - c * k), 1) - s * c * k / 3 * RD(ABS(1 - c), ABS(1 - c * k), 1);
 E *= (7 * r * r + delta * delta - 4) * 2 * sqq;
 E /= 9;
 res += E;
 if(r > delta) res += 2.0943951023931954923L;
 if(r == delta) res += 1.0471975511965977462L;
 return res;
}


long double D1L1(long double delta, long double r)
{
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double rmindeltaq = rmindelta * rmindelta, rdelta4 = 4 * delta * r, k, s, c, F, E;
 if (ABS(rmindelta) >= 1 || r <= 0 || delta <= 0) return 0;
 k = rdelta4 / (1 - rmindeltaq);
 s = (k < 1) ? 1 : 1 / sqrtl(k);
 c = (k < 1) ? 1 : 1 / k;
 F = s * RF(ABS(1 - c), ABS(1 - c * k), 1);
 F *= 2 * (rplusdelta * rplusdelta - 1);
 E = s * RF(ABS(1 - c), ABS(1 - c * k), 1) - s * c * k / 3 * RD(ABS(1 - c), ABS(1 - c * k), 1);
 E *= 2 * (1- r * r - delta * delta);
 return -(E + F) * sqrtl(1 - rmindeltaq) / delta / 3;
}

long double D2L1(long double delta, long double r)
{
 long double rmindelta = r - delta;
 long double rmindeltaq = rmindelta * rmindelta, rdelta4 = 4 * delta * r, k, s, c, E;
 if (ABS(rmindelta) >= 1 || r <= 0 || delta < 0) return 0;
 k = rdelta4 / (1 - rmindeltaq);
 s = (k < 1) ? 1 : 1 / sqrtl(k);
 c = (k < 1) ? 1 : 1 / k;
 E = s * RF(ABS(1 - c), ABS(1 - c * k), 1) - s * c * k / 3 * RD(ABS(1 - c), ABS(1 - c * k), 1);
 return 4 * r * E * sqrtl(1 - rmindeltaq);
}


long double L2(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi1 = (delta * delta + 1 - r * r) / 2;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
 {
  psi1 = (-2 * SIGN(psi1) + 1) * PI_2;
  psi2 = (-2 * SIGN(psi2) + 1) * PI_2;
 }
 else
 {
  psi1 = ReArcCos(psi1 / delta);
  psi2 = ReArcCos(psi2 / delta);
 }
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return  psi1 / 2 + psi2 * r * r * (2 * delta * delta + r * r) / 2 - Q * (delta * delta + 5 * r * r + 1) / 8;
}

long double D1L2(long double delta, long double r)
{
 if(r <= 0 || delta <= 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 psi2 = ReArcCos(psi2 / delta);
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return 2 * psi2 * delta * r * r  - (delta * delta + r * r + 1) / 2 / delta * Q;
}

long double D2L2(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
  psi2 = (-2 * SIGN(psi2) + 1) * PI_2;
 else
  psi2 = ReArcCos(psi2 / delta);
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return 2 * r * ((delta * delta + r * r) * psi2 - Q) ;
}

long double L3(long double delta, long double r)
{
    int i;
    long double X0, X1, sx, s = 0, t, ts, dmrsq, dmr = delta - r, dr4, n, k;
    dmrsq = dmr * dmr;
    if(ABS(dmrsq) >= 1.0L || r <= 0.0L || delta < 0.0L)
        return 0;
    dr4 = 4 * delta * r;    
    n = dmrsq;    
    k = (1 - dmrsq);    
    if(k <= 0)
        return 0;
    if(k > dr4)
        X1 = PI_2;
    if(k < dr4)
    {           
        X0 = asinl(sqrtl(k / dr4));
        X1 = X0 + X0 - PI_2;
        X1 = MAX(X1, X0 - 0.5L);
        X1 = MAX(0, X1);                        
        t = k - n;
        if (t > 0)            
            t = asinl(sqrtl(t / dr4 / 2));
        else 
            t = 0;
        X1 = MAX(t, X1);        
        for(i = 0;i < nodes_N; ++i)
        {
            t = X1 + nodes_Jacobi1d4[i] * (X0 - X1);
            dmrsq = sinl(t);
            dmrsq *= dmrsq;
            ts = dmrsq * dr4;            
            t = k - ts;
            s += (dmr - 2 * delta * dmrsq) * t * sqrtl(sqrtl(t / (1.0L - nodes_Jacobi1d4[i]))) / (n + ts) * weights_Jacobi1d4[i];
        }
        s *= (X0 - X1);        
    }
    if(k == dr4)
    {
        X0 = PI_2;
        X1 = MAX(X1, X0 - 0.5L);
        X1 = MAX(0, X1);  
        t = k - n;
        if (t > 0)
            t = asinl(sqrtl(t / 2 / dr4));
        else            
            t = 0;
        X1 = MAX(t, X1); 
        dmrsq = dr4 * sqrtl(sqrtl(dr4));
        for(i = 0;i < nodes_N; ++i)
        {
            t = X1 + nodes_Jacobi1d2[i] * (X0 - X1);
            sincosl(t, &ts, &t);
            ts *= ts * dr4;
            s += dmrsq * (dmr - 2 * delta * ts) * t * t * sqrtl(t / (1.0L - nodes_Jacobi1d2[i])) / (n + ts) * weights_Jacobi1d2[i];
        }
        s *= (X0 - X1);
    }
    do
    {
        X0 = X1;
        X1 = X0 - 0.5L;
        ts = sinl(X0);
        ts *= ts * dr4;
        t = 2 * ts - k;
        if (t > 0)
            t = asinl(sqrtl(t / dr4));
        else
            t = 0;        
        X1 = MAX(t, X1);
        if(n > 0.0L)
        {
            t = ts - n;
            if (t > 0)
                t = asinl(sqrtl(t / 2 / dr4));
            else 
                t = 0;
            X1 = MAX(t, X1);
        }
        for(i = 0, sx = 0;i < nodes_N; ++i)
        {
            t = X1 + nodes_Legendre[i] * (X0 - X1);
            dmrsq = sinl(t);
            dmrsq *= dmrsq;
            ts = dmrsq * dr4;            
            t = k - ts;
            sx += (dmr - 2 * delta * dmrsq) * t * sqrtl(sqrtl(t)) / (n + ts) * weights_Legendre[i];
        }
        s += sx * (X0 - X1);
    }while(X1 > 0);
    s *= 1.6L * r;
    if(r > delta)
        s += 2.51327412287183459077L;
    if(r == delta)
        s += 1.256637061435917295385L;        
    return s;
}

long double D1L3(long double delta, long double r)
{
    int i;
    long double X0, X1, sx, s = 0, t, ts, tc, c2x, c2, dmr = delta - r, dr4, k;
    if(ABS(dmr) >= 1.0L || r <= 0.0L || delta < 0.0L)
        return 0;
    dr4 = 4 * delta * r;
    k = (1 - dmr * dmr) / dr4;
    if(k <= 0)
        return 0;
    if(k > 1)
        X1 = PI_2;
    if(k < 1)
    {           
        X0 = asinl(sqrtl(k));
        X1 = X0 + X0 - PI_2;
        X1 = MAX(X1, X0 - 0.5L);
        X1 = MAX(0, X1);                                
        for(i = 0, sx = 0;i < nodes_N; ++i)
        {
            ts = X1 + nodes_Jacobi1d4[i] * (X0 - X1);
            c2x = cosl(ts * 2);
            ts = sinl(ts);
            ts *= ts;            
            ts = k - ts;
            s += c2x * sqrtl(sqrtl(ts / (1.0L - nodes_Jacobi1d4[i]))) * weights_Jacobi1d4[i];
        }
        s *= (X0 - X1);        
    }
    if(k == 1)    
        return -4 * r * sqrtl(sqrtl(dr4)) * 0.39938007824519740247997L;
    if(k > 1.5)    
        do
        {
            
            X0 = X1;
            X1 = X0 - 0.5L;
            ts = sinl(X0);
            ts *= ts;
            ts = 2 * ts - k;
            if (ts > 0)
                ts = asinl(sqrtl(ts));
            else
                ts = 0;        
            X1 = MAX(ts, X1);
            for(i = 0, sx = 0;i < nodes_N; ++i)
            {
                t = X1 + nodes_Legendre[i] * (X0 - X1);
                c2x = cosl(t * 2); 
                c2 = c2x / 2;
                t = k - 0.5L;
                ts = sqrtl(t + c2);
                tc = sqrtl(t - c2);                
                c2x = c2x * c2;                
                c2x /= ts + tc;
                ts = sqrtl(ts);
                tc = sqrtl(tc);
                c2x /= ts + tc;
                sx += c2x * weights_Legendre[i];
            }
            s += sx * (X0 - X1);
        }while(X1 > 0);            
    else    
        do
        {
            X0 = X1;
            X1 = X0 - 0.5L;
            ts = sinl(X0);
            ts *= ts;
            ts = 2 * ts - k;
            if (ts > 0)
                ts = asinl(sqrtl(ts));
            else
                ts = 0;        
            X1 = MAX(ts, X1);
            for(i = 0, sx = 0;i < nodes_N; ++i)
            {
                ts = X1 + nodes_Legendre[i] * (X0 - X1);
                c2x = cosl(ts * 2);                
                ts = sinl(ts);
                ts *= ts;            
                ts = k - ts;
                sx += c2x * sqrtl(sqrtl(ts)) * weights_Legendre[i];
            }
            s += sx * (X0 - X1);
        }while(X1 > 0);
    return -4 * r * sqrtl(sqrtl(dr4)) * s;    
}

long double D2L3(long double delta, long double r)
{
    int i;
    long double X0, X1, sx, s = 0, t, dmr = delta - r, dr4, k;
    if(ABS(dmr) >= 1.0L || r <= 0.0L || delta < 0.0L)
        return 0;
    dr4 = 4 * delta * r;
    k = (1 - dmr * dmr) / dr4;
    if(k <= 0)
        return 0;
    if(k > 1)
        X1 = PI_2;
    if(k < 1)
    {           
        X0 = asinl(sqrtl(k));
        X1 = X0 + X0 - PI_2;
        X1 = MAX(X1, X0 - 0.5L);
        X1 = MAX(0, X1);
        for(i = 0, sx = 0;i < nodes_N; ++i)
        {
            t = X1 + nodes_Jacobi1d4[i] * (X0 - X1);
            t = sinl(t);
            t *= t;            
            t = k - t;
            s += sqrtl(sqrtl(t / (1.0L - nodes_Jacobi1d4[i]))) * weights_Jacobi1d4[i];
        }
        s *= (X0 - X1);        
    }
    if(k == 1)    
       return 4 * r * sqrtl(sqrtl(dr4)) * 0.71888414084135532446395L;    
    do
    {
        X0 = X1;
        X1 = X0 - 0.5L;        
        t = sinl(X0);
        t *= t;
        t = 2 * t - k;
        if(t > 0)
            t = asinl(sqrtl(t));
        else
            t = 0;        
        X1 = MAX(t, X1);        
        for(i = 0, sx = 0;i < nodes_N; ++i)
        {
            t = X1 + nodes_Legendre[i] * (X0 - X1);
            t = sinl(t);
            t *= t;
            t = k - t;
            sx += sqrtl(sqrtl(t)) * weights_Legendre[i];
        }
        s += sx * (X0 - X1);
    }while(X1 > 0);
    return 4 * r * sqrtl(sqrtl(dr4)) * s;
}

long double Kepler(long double x, long double e)
{    
    int cnt = 0;
    //int cnt2 = 0;
    bool neg_e = false, adjust = false;
    long double x1, x2, s, c, d1, d2, d, a, b, v;
    if(fabsl(e) > 1.0L)
        return 0.0L;    
    if(sinl(x) == 0.0L)
        return x;    
    if(e < 0)
    {
        e = -e;
        x -= PI;        
        neg_e = true;
    }
    a = x - e;
    b = x + e;
    if(sinl(x / 2) > 0)    
        x1 = MAX(a, (long double)(lroundl((x - PI) / (2 * PI))) * (2 * PI) + PI);
    else
        x1 = MIN(b, (long double)(lroundl((x - PI) / (2 * PI))) * (2 * PI) + PI);
    sincosl(x1, &s, &c);
    v = (1 - e * c);
    d1 = (x1 - e * s - x) / v;    
    do
    {   
        x2 = x1;
        d2 = d1;        
        x1 -= d1;
        sincosl(x1 , &s, &c);
        d1 = (x1 - e * s - x);        

        if(d1 == 0.0L)
            break;        

        if((d1 < 0) == (d2 > 0))
        {
            ++cnt;
            if(cnt > 1)
            {

                adjust = true;
                break;
            }
        }        
        if(cnt == 0)
            v = (1 - e * c);
        d1 /= v;
    }while(fabsl((x1 - x2) / x2) > 2 * LDBL_EPSILON);

    if(adjust)
    {
        do
        {  
            //++cnt2;
            c = (x1 + x2) / 2;
            d = c - e * sinl(c) - x;
            if(d == 0)
            {
                x1 = c;
                break;
            }
            if((d < 0) == (d1 > 0))
            {
                x2 = c;
                d2 = d;
            }
            else
            {
                x1 = c;
                d1 = d;
            }
        }while(fabsl((x1 - x2) / x2) > LDBL_EPSILON);
        x1 = c;
    }    

    if(neg_e) return x1 + PI;
    else return x1;
}





long double DeltaEccentric_omega(long double fi, long double e, long double omega, long double i)
{
    long double s, c, nu, ci;
    fi = Kepler(fi, e);
    sincosl(fi / 2, &s, &c);
    nu = 2 * atanl(s / c * sqrtl((1 + e) / (1 - e)));
    sincosl(i, &s, &ci);
    c = cosl(fi);
    fi = cosl(nu + omega);
    return (1 - e * c) * sqrtl(ci * ci + s * s * fi * fi);
}

int DeltaEccentric_omega_d(long double fi, long double e, long double omega, long double i, long double *delta, long double *df, long double *de, long double *dw, long double *di)
{
    long double s, c, nu, si, ci, sqr, ec1, sinnuw, sin2nuw, sqr1ec;    
    fi = Kepler(fi, e);
    sincosl(fi / 2, &s, &c);
    nu = 2 * atanl(s / c * sqrtl((1 + e) / (1 - e)));
    sincosl(fi, &s, &c);
    sincosl(i, &si, &ci);    
    ec1 = nu + omega;
    nu = sqrtl(1 - e * e);
    sincosl(ec1, &sinnuw, &fi);    
    sin2nuw = sinl(2 * ec1);
    ec1 = 1 - e * c;
    si *= si;
    sqr = sqrtl(ci * ci + si * fi * fi);
    *delta = sqr * ec1;    
    sqr1ec = sqr / ec1;
    *dw = si * sin2nuw;
    *de = (e - c) * sqr1ec - (*dw) * s / (sqr * 2) * (1 / nu + nu / ec1);    
    *df = e * s * sqr1ec - nu * (*dw) / (*delta) / 2;
    sqr1ec *= -2;
    *dw /= sqr1ec;
    *di = sinnuw * sinnuw * sinl(2 * i) / sqr1ec;
    return (sinnuw < 0);
}

int DeltaEccentric_omega_ci_d(long double fi, long double e, long double omega, long double cisqrt, long double *delta, long double *df, long double *de, long double *dw, long double *di)
{
    long double s, c, nu, sqr, ec1, sinnuw, sin2nuw, sqr1ec;
    int res;
    fi = Kepler(fi, e);
    sincosl(fi / 2, &s, &c);
    nu = 2 * atanl(s / c * sqrtl((1 + e) / (1 - e)));
    sincosl(fi, &s, &c);    
    ec1 = nu + omega;
    nu = sqrtl(1 - e * e);
    sincosl(ec1, &sinnuw, &fi);    
    sin2nuw = sinl(2 * ec1);
    ec1 = 1 - e * c;   
    cisqrt -= 1;
    res = (sinnuw < 0);
    sinnuw *= sinnuw;
    sqr = sqrtl(cisqrt * sinnuw + 1 );
    *delta = sqr * ec1;    
    sqr1ec = sqr / ec1;
    *dw = - cisqrt * sin2nuw;
    *de = (e - c) * sqr1ec - (*dw) * s / (sqr * 2) * (1 / nu + nu / ec1);    
    *df = e * s * sqr1ec - nu * (*dw) / (*delta) / 2;
    sqr1ec *= -2;
    *dw /= sqr1ec;
    *di = - sinnuw / sqr1ec;
    return res;
}

void DeltaEccentric_omega_d_ext(long double fi, long double e, long double omega, long double i, long double *delta, long double *df, long double *de, long double *dw, long double *di, long double *dfw)
{
    long double s, c, nu, si, ci, ec, sqr, ec1, sinnuw, sin2nuw, sqr1ec;
    if(e >= 1.0L)
        return;
    fi = Kepler(fi, e);
    sincosl(fi / 2, &s, &c);
    nu = 2 * atanl(s / c * sqrtl((1 + e) / (1 - e)));
    sincosl(fi, &s, &c);
    sincosl(i, &si, &ci);    
    ec1 = nu + omega;
    nu = sqrtl(1 - e * e);
    sincosl(ec1, &sinnuw, &fi);
    sin2nuw = sinl(2 * ec1);
    ec = e * c;
    ec1 = 1 - ec;
    *di = ec1 * ec1;
    si *= si;
    sqr = sqrtl(ci * ci + si * fi * fi);
    *delta = sqr * ec1;
    sqr1ec = sqr / ec1;
    *dw = si * sin2nuw;
    *df = (*dw) / (sqr * 2);
    *de = (e - c) * sqr1ec - (*df) * s * (1 / nu + nu / ec1);    
    *dfw = (s * sqr - (*df) * (-e - c * (ec - 2) * (1 + (*di))) / (nu + (*di))) / ec1 * e;
    *df = e * s * sqr1ec - nu * (*dw) / (*delta) / 2;
    sqr1ec *= -2;
    *dw /= sqr1ec;
    *di = sinnuw * sinnuw * sinl(2 * i) / sqr1ec;
}

long double phmin(long double e, long double w, long double i)
{
	long double f1, f2, s, c, mu, se, sqe = sqrtl(1 - e * e);
	sincosl(3 * PI_4 - w / 2, &s, &c);
	f1 = 2 * atanl(sqrtl((1 - e) / (1 + e)) * s / c);
        f1 = f1 - e * sinl(f1);
	do
	{
		f2 = f1;
                sincosl(i , &s, &c);
		f1 = c / s;
                f1 *= f1;
                se = Kepler(f2, e);
                sincosl(se / 2, &s, &c);
                mu = 2 * atanl(sqrtl((1 + e) / (1 - e)) * s / c);
                sincosl(mu + w, &s, &c);
                mu = e * sinl(se);
                f1 = PI + acosl(mu * f1 / (mu * c - sqe * s)) - w;
                sincosl(f1 / 2, &s, &c);
                f1 = 2 * atanl(sqrtl((1 - e) / (1 + e)) * s / c);
                f1 = f1 - e * sinl(f1);
	}while(fabsl((f1 - f2) / f2) >=  LDBL_EPSILON);
	return f1;
}

long double phmin_ci(long double e, long double w, long double ci)
{
	long double f1, f2, s, c, mu, se, sqe = sqrtl(1 - e * e);
	sincosl(3 * PI_4 - w / 2, &s, &c);
	f1 = 2 * atanl(sqrtl((1 - e) / (1 + e)) * s / c);
        f1 = f1 - e * sinl(f1);
	do
	{
		f2 = f1;                
		f1 = ci / (1 - ci);                
                se = Kepler(f2, e);
                sincosl(se / 2, &s, &c);
                mu = 2 * atanl(sqrtl((1 + e) / (1 - e)) * s / c);
                sincosl(mu + w, &s, &c);
                mu = e * sinl(se);
                f1 = PI + acosl(mu * f1 / (mu * c - sqe * s)) - w;
                sincosl(f1 / 2, &s, &c);
                f1 = 2 * atanl(sqrtl((1 - e) / (1 + e)) * s / c);
                f1 = f1 - e * sinl(f1);
	}while(fabsl((f1 - f2) / f2) >=  LDBL_EPSILON);
	return f1;
}

long double Lustre(long double fi, long double Rs, long double Rp, long double e, long double w, long double i, long double koeff[6])
{
    long double delta = DeltaEccentric_omega(fi, e, w, i) / Rs, r = Rp / Rs;
    long double Ldec0 = PI - L0(delta, r), Ldec1 = 2 * PI / 3 - L1(delta, r), Ldec2 = PI_2 - L2(delta, r), Ldec3 = 4 * PI / 5 - L3(delta, r);
    long double Ldecl = Ldec1 - Ldec0;
    return Rs * Rs * (koeff[0] * Ldec0 + koeff[1] * Ldecl + koeff[2] * (2 * Ldecl + Ldec2) + koeff[3] * (Ldec3 - Ldec0));
}


long double Lustre_norm(long double fi, long double e, long double w, long double i, long double Rs, long double Rp, long double koeff[5])
{
    long double delta = DeltaEccentric_omega(fi, e, w, i) / Rs, r = Rp / Rs;
    long double Ldec0 = PI - L0(delta, r), Ldec1 = 2 * PI / 3 - L1(delta, r), Ldec2 = PI_2 - L2(delta, r), Ldec3 = 4 * PI / 5 - L3(delta, r);
    long double Ldecl = Ldec1 - Ldec0;
    return (Ldec0 + koeff[0] * Ldecl + koeff[1] * (2 * Ldecl + Ldec2) + koeff[2] * (Ldec3 - Ldec0)) / (1 - koeff[0] / 3 - koeff[1] / 6 - koeff[2] / 5) / PI;
}


